from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site

from djoser import utils
from django.conf import settings as django_settings
from djoser.conf import settings
from django.core import mail
from django.template.context import make_context
from django.template.loader import get_template
from django.views.generic.base import ContextMixin


class BaseEmailMessage(mail.EmailMultiAlternatives, ContextMixin):
    _node_map = {
        "subject": "subject",
        "text_body": "body",
        "html_body": "html",
    }
    template_name = None

    def __init__(self, request=None, context=None, template_name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.request = request
        self.context = {} if context is None else context
        self.html = None

        if template_name is not None:
            self.template_name = template_name

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        context = dict(ctx, **self.context)
        if self.request:
            site = get_current_site(self.request)
            domain = context.get("domain") or (
                getattr(django_settings, "DOMAIN", "") or site.domain
            )
            protocol = context.get("protocol") or (
                "https" if self.request.is_secure() else "http"
            )
            site_name = context.get("site_name") or (
                getattr(django_settings, "SITE_NAME", "") or site.name
            )
            user = context.get("user") or self.request.user
        else:
            domain = context.get("domain") or getattr(django_settings, "DOMAIN", "")
            protocol = context.get("protocol") or "http"
            site_name = context.get("site_name") or getattr(
                django_settings, "SITE_NAME", ""
            )
            user = context.get("user")

        context.update(
            {
                "domain": domain,
                "protocol": protocol,
                "site_name": site_name,
                "user": user,
            }
        )
        return context

    def render(self):
        context = make_context(self.get_context_data(), request=self.request)
        template = get_template(self.template_name)
        with context.bind_template(template.template):
            for node in template.template.nodelist:
                self._process_node(node, context)
        self._attach_body()

    # custom interface incompatible with django, `to` is a required param
    def send(self, to, fail_silently=False, **kwargs):
        self.render()

        self.to = to
        self.cc = kwargs.pop("cc", [])
        self.bcc = kwargs.pop("bcc", [])
        self.reply_to = kwargs.pop("reply_to", [])
        self.from_email = kwargs.pop("from_email", django_settings.DEFAULT_FROM_EMAIL)
        self.request = None
        super().send(fail_silently=fail_silently)

    def _process_node(self, node, context):
        attr = self._node_map.get(getattr(node, "name", ""))
        if attr is not None:
            setattr(self, attr, node.render(context).strip())

    def _attach_body(self):
        if self.body and self.html:
            self.attach_alternative(self.html, "text/html")
        elif self.html:
            self.body = self.html
            self.content_subtype = "html"


class BaseDjoserEmail(BaseEmailMessage):
    def get_context_data(self):
        context = super().get_context_data()
        overridable = {
            "protocol": settings.EMAIL_FRONTEND_PROTOCOL,
            "domain": settings.EMAIL_FRONTEND_DOMAIN,
            "site_name": settings.EMAIL_FRONTEND_SITE_NAME,
        }
        for context_key, context_value in overridable.items():
            if context_value:
                context.update({context_key: context_value})
        context.pop("view", None)
        return context


class ActivationEmail(BaseDjoserEmail):
    template_name = "email/activation.html"

    def get_context_data(self):
        # ActivationEmail can be deleted
        context = super().get_context_data()

        user = context.get("user")
        context["uid"] = utils.encode_uid(user.pk)
        context["token"] = default_token_generator.make_token(user)
        context["url"] = settings.ACTIVATION_URL.format(**context)
        return context


class ConfirmationEmail(BaseDjoserEmail):
    template_name = "email/confirmation.html"


class PasswordResetEmail(BaseDjoserEmail):
    template_name = "email/password_reset.html"

    def get_context_data(self):
        # PasswordResetEmail can be deleted
        context = super().get_context_data()

        user = context.get("user")
        context["uid"] = utils.encode_uid(user.pk)
        context["token"] = default_token_generator.make_token(user)
        context["url"] = settings.PASSWORD_RESET_CONFIRM_URL.format(**context)
        return context


class PasswordChangedConfirmationEmail(BaseDjoserEmail):
    template_name = "email/password_changed_confirmation.html"


class UsernameChangedConfirmationEmail(BaseDjoserEmail):
    template_name = "email/username_changed_confirmation.html"


class UsernameResetEmail(BaseDjoserEmail):
    template_name = "email/username_reset.html"

    def get_context_data(self):
        context = super().get_context_data()

        user = context.get("user")
        context["uid"] = utils.encode_uid(user.pk)
        context["token"] = default_token_generator.make_token(user)
        context["url"] = settings.USERNAME_RESET_CONFIRM_URL.format(**context)
        return context



from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.views import APIView
from webauthn import (
    WebAuthnAssertionOptions,
    WebAuthnAssertionResponse,
    WebAuthnMakeCredentialOptions,
    WebAuthnRegistrationResponse,
    WebAuthnUser,
)
from webauthn.webauthn import (
    AuthenticationRejectedException,
    RegistrationRejectedException,
)

from djoser import signals
from djoser.compat import get_user_email
from djoser.conf import settings
from djoser.utils import login_user

from .models import CredentialOptions
from .serializers import WebauthnLoginSerializer, WebauthnSignupSerializer
from .utils import create_challenge

User = get_user_model()


class SignupRequestView(APIView):
    permission_classes = (AllowAny,)
    serializer_class = WebauthnSignupSerializer

    def post(self, request):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        co = serializer.save()

        credential_registration_dict = WebAuthnMakeCredentialOptions(
            challenge=co.challenge,
            rp_name=settings.WEBAUTHN["RP_NAME"],
            rp_id=settings.WEBAUTHN["RP_ID"],
            user_id=co.ukey,
            username=co.username,
            display_name=co.display_name,
            icon_url="",
        )

        return Response(credential_registration_dict.registration_dict)


class SignupView(APIView):
    permission_classes = (AllowAny,)
    serializer_class = settings.WEBAUTHN.SIGNUP_SERIALIZER

    def post(self, request, ukey):
        co = get_object_or_404(CredentialOptions, ukey=ukey)
        user_serializer = self.serializer_class(data=request.data)
        user_serializer.is_valid(raise_exception=True)

        webauthn_registration_response = WebAuthnRegistrationResponse(
            rp_id=settings.WEBAUTHN["RP_ID"],
            origin=settings.WEBAUTHN["ORIGIN"],
            registration_response=request.data,
            challenge=co.challenge,
            none_attestation_permitted=True,
        )
        try:
            webauthn_credential = webauthn_registration_response.verify()
        except RegistrationRejectedException:
            return Response(
                {api_settings.NON_FIELD_ERRORS_KEY: "WebAuthn verification failed."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        user = user_serializer.save()
        co.challenge = ""
        co.user = user
        co.sign_count = webauthn_credential.sign_count
        co.credential_id = webauthn_credential.credential_id.decode()
        co.public_key = webauthn_credential.public_key.decode()
        co.save()
        signals.user_registered.send(
            sender=self.__class__, user=user, request=self.request
        )

        if settings.SEND_ACTIVATION_EMAIL and not user.is_active:
            context = {"user": user}
            to = [get_user_email(user)]
            settings.EMAIL.activation(self.request, context).send(to)

        return Response(user_serializer.data, status=status.HTTP_201_CREATED)


class LoginRequestView(APIView):
    permission_classes = (AllowAny,)
    serializer_class = WebauthnLoginSerializer

    def post(self, request):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        co = CredentialOptions.objects.get(
            username=serializer.validated_data["username"]
        )

        co.challenge = create_challenge(32)
        co.save()

        webauthn_user = WebAuthnUser(
            user_id=co.ukey,
            username=co.username,
            display_name=co.display_name,
            icon_url="",
            credential_id=co.credential_id,
            public_key=co.public_key,
            sign_count=co.sign_count,
            rp_id=settings.WEBAUTHN["RP_ID"],
        )
        webauthn_assertion_options = WebAuthnAssertionOptions(
            webauthn_user, co.challenge
        )

        return Response(webauthn_assertion_options.assertion_dict)


# this name looks good :)
class LoginView(APIView):
    permission_classes = (AllowAny,)
    serializer_class = settings.WEBAUTHN.LOGIN_SERIALIZER

    def post(self, request):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.user
        co = user.credential_options

        webuathn_user = WebAuthnUser(
            user_id=co.ukey,
            username=user.username,
            display_name=co.display_name,
            icon_url="",
            credential_id=co.credential_id,
            public_key=co.public_key,
            sign_count=co.sign_count,
            rp_id=settings.WEBAUTHN["RP_ID"],
        )

        webauthn_assertion_response = WebAuthnAssertionResponse(
            webuathn_user,
            request.data,
            co.challenge,
            settings.WEBAUTHN["ORIGIN"],
            uv_required=False,
        )

        try:
            sign_count = webauthn_assertion_response.verify()
        except AuthenticationRejectedException:
            return Response(
                {api_settings.NON_FIELD_ERRORS_KEY: "WebAuthn verification failed."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        co.sign_count = sign_count
        co.challenge = ""
        co.save()

        token_serializer_class = settings.SERIALIZERS.token
        token = login_user(request, user)
        return Response(
            token_serializer_class(token).data, status=status.HTTP_201_CREATED
        )



from django.contrib.auth import get_user_model
from rest_framework.routers import DefaultRouter

from djoser import views

router = DefaultRouter()
router.register("users", views.UserViewSet)

User = get_user_model()

urlpatterns = router.urls



import string
from random import SystemRandom

random = SystemRandom()
challenge_characters = string.ascii_letters + string.digits


def create_challenge(length):
    return "".join(random.choices(challenge_characters, k=length))


def create_ukey(length):
    return create_challenge(length)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djoser',
 'djoser.social',
 'djoser.social.backends',
 'djoser.social.token',
 'djoser.urls',
 'djoser.webauthn',
 'djoser.webauthn.migrations']

package_data = \
{'': ['*'],
 'djoser': ['locale/ca/LC_MESSAGES/*',
            'locale/de/LC_MESSAGES/*',
            'locale/es/LC_MESSAGES/*',
            'locale/fr/LC_MESSAGES/*',
            'locale/id/LC_MESSAGES/*',
            'locale/ja/LC_MESSAGES/*',
            'locale/ka/LC_MESSAGES/*',
            'locale/me/LC_MESSAGES/*',
            'locale/pl/LC_MESSAGES/*',
            'locale/pt_BR/LC_MESSAGES/*',
            'locale/ru_RU/LC_MESSAGES/*',
            'templates/email/*']}

install_requires = \
['django>=3.0.0',
 'djangorestframework-simplejwt>=5.0,<6.0',
 'social-auth-app-django>=5.0.0,<6.0.0']

extras_require = \
{'djet': ['djet>=0.3.0,<0.4.0'], 'webauthn': ['webauthn<1.0']}

setup_kwargs = {
    'name': 'djoser',
    'version': '2.3.1',
    'description': 'REST implementation of Django authentication system.',
    'long_description': '======\ndjoser\n======\n\n.. image:: https://img.shields.io/pypi/v/djoser.svg\n   :target: https://pypi.org/project/djoser\n\n.. image:: https://github.com/sunscrapers/djoser/workflows/Tests/badge.svg\n    :target: https://github.com/sunscrapers/djoser/actions?query=branch%3Amaster+workflow%Tests++\n    :alt: Build Status\n\n.. image:: https://codecov.io/gh/sunscrapers/djoser/branch/master/graph/badge.svg\n :target: https://codecov.io/gh/sunscrapers/djoser\n\n.. image:: https://img.shields.io/pypi/dm/djoser\n   :target: https://img.shields.io/pypi/dm/djoser\n\n.. image:: https://readthedocs.org/projects/djoser/badge/?version=latest\n    :target: https://djoser.readthedocs.io/en/latest/\n    :alt: Docs\n\nREST implementation of `Django <https://www.djangoproject.com/>`_ authentication\nsystem. **djoser** library provides a set of `Django Rest Framework <https://www.django-rest-framework.org/>`_\nviews to handle basic actions such as registration, login, logout, password\nreset and account activation. It works with\n`custom user model <https://docs.djangoproject.com/en/dev/topics/auth/customizing/>`_.\n\nInstead of reusing Django code (e.g. ``PasswordResetForm``), we reimplemented\nfew things to fit better into `Single Page App <https://en.wikipedia.org/wiki/Single-page_application>`_\narchitecture.\n\nDeveloped by `SUNSCRAPERS <http://sunscrapers.com/>`_ with passion & patience.\n\n.. image:: https://asciinema.org/a/94J4eG2tSBD2iEfF30a6vGtXw.png\n  :target: https://asciinema.org/a/94J4eG2tSBD2iEfF30a6vGtXw\n\nRequirements\n============\n\nTo be able to run **djoser** you have to meet the following requirements:\n\n- Python>=3.8\n- Django>=3.0.0\n- Django REST Framework>=3.12\n\nInstallation\n============\n\nSimply install using ``pip``:\n\n.. code-block:: bash\n\n    $ pip install djoser\n\nAnd continue with the steps described at\n`configuration <https://djoser.readthedocs.io/en/latest/getting_started.html#configuration>`_\nguide.\n\nDocumentation\n=============\n\nDocumentation is available to study at\n`https://djoser.readthedocs.io <https://djoser.readthedocs.io>`_\nand in ``docs`` directory.\n\nContributing and development\n============================\n\nTo start developing on **djoser**, clone the repository:\n\n.. code-block:: bash\n\n    $ git clone git@github.com:sunscrapers/djoser.git\n\nWe use `poetry <https://python-poetry.org/>`_ as dependency management and packaging tool.\n\n.. code-block:: bash\n\n    $ cd djoser\n    $ poetry install --all-extras\n\nThis will create a virtualenv with all development dependencies.\n\nTo run the test just type:\n\n.. code-block:: bash\n\n    $ poetry run py.test testproject\n\nWe also prepared a convenient ``Makefile`` to automate commands above:\n\n.. code-block:: bash\n\n    $ make init\n    $ make test\n\nTo activate the virtual environment run\n\n.. code-block:: bash\n\n    $ poetry shell\n\nWithout poetry\n--------------\n\nNew versions of ``pip`` can use ``pyproject.toml`` to build the package and install its dependencies.\n\n.. code-block:: bash\n\n    $ pip install .[test]\n\n.. code-block:: bash\n\n    $ cd testproject\n    $ ./manage.py test\n\nExample project\n---------------\n\nYou can also play with test project by running following commands:\n\n.. code-block:: bash\n\n    $ make migrate\n    $ make runserver\n\nCommiting your code\n-------------------\n\nBefore sending patches please make sure you have `pre-commit <https://pre-commit.com/>`_ activated in your local git repository:\n\n.. code-block:: bash\n\n    $ pre-commit install\n\nThis will ensure that your code is cleaned before you commit it.\n\nSimilar projects\n================\n\nList of projects related to Django, REST and authentication:\n\n- `django-rest-registration <https://github.com/apragacz/django-rest-registration>`_\n- `django-oauth-toolkit <https://github.com/evonove/django-oauth-toolkit>`_\n\nPlease, keep in mind that while using custom authentication and TokenCreateSerializer\nvalidation, there is a path that **ignores intentional return of None** from authenticate()\nand try to find User using parameters. Probably, that will be changed in the future.\n',
    'author': 'Sunscrapers',
    'author_email': 'info@sunscrapers.com',
    'maintainer': 'Tomasz Wojcik',
    'maintainer_email': 'djoser@tomwojcik.com',
    'url': 'https://github.com/sunscrapers/djoser',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
